﻿#region License Information

/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */

#endregion

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using HeuristicLab.Analysis;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.DatastreamAnalysis {
  internal enum DatastreamAnalysisOptimizerAction {
    None,
    Prepare,
    Start,
    Stop,
    Pause
  }

  [StorableClass]
  [Item("DatastreamAnalysis Optimizer",
     "The main loop for evaluating ensemble models against a incoming datastream of time series fashion.")]
  [Creatable(CreatableAttribute.Categories.Algorithms)]
  public class DatastreamAnalysisOptimizer : Executable, IOptimizer, IStorableContent {
    #region properties
    public string Filename { get; set; }

    private DatastreamAnalysisOptimizerAction daoAction;

    public IEnumerable<IOptimizer> NestedOptimizers { get; }


    [Storable]
    protected ILog log;
    public ILog Log {
      get { return log; }
    }

    [Storable]
    private ResultCollection results;

    public ResultCollection Results {
      get { return results; }
    }

    private CancellationTokenSource cancellationTokenSource;
    private bool stopPending;
    private DateTime lastUpdateTime;

    [Storable]
    protected int runsCounter;

    [Storable]
    private RunCollection runs;

    public RunCollection Runs
    {
      get { return runs; }
      protected set
      {
        if (value == null) throw new ArgumentNullException();
        if (runs != value) {
          if (runs != null) DeregisterRunsEvents();
          runs = value;
          if (runs != null) RegisterRunsEvents();
        }
      }
    }


    [Storable]
    private IItemList<RegressionEnsembleModel> ensembles;

    public IItemList<RegressionEnsembleModel> Ensembles {
      get { return ensembles; }
      set {
        if (value == null || value == ensembles)
          return;

        ensembles = value;
      }
    }


    // VAR 1: datastream ~= problem data, VAR 2 (TODO): datastream = external source e.g. webservice, AMQP-Queue, etc.
    [Storable]
    private Datastream datastream;

    public Datastream Datastream {
      get { return datastream; }
      set {
        if (value == null || value == datastream)
          return;

        datastream = value;
      }
    }

    #endregion properties

    #region ResultsProperties
    private double ResultsBestQuality
    {
      get { return ((DoubleValue)Results["Best Quality"].Value).Value; }
      set { ((DoubleValue)Results["Best Quality"].Value).Value = value; }
    }
    private DataTable ResultsQualities
    {
      get { return ((DataTable)Results["Qualities"].Value); }
    }
    #endregion

    #region constructors, cloner,...

    public DatastreamAnalysisOptimizer() {
      name = "Datastream Analysis";
      log = new Log();
      results = new ResultCollection();
      ensembles = new ItemList<RegressionEnsembleModel>();
      datastream = new Datastream();
      runsCounter = 0;
      runs = new RunCollection();
      Initialize();
    }

    [StorableConstructor]
    protected DatastreamAnalysisOptimizer(bool deserializing) : base(deserializing) {
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      // nothing to do in here
    }

    protected DatastreamAnalysisOptimizer(DatastreamAnalysisOptimizer original, Cloner cloner) : base(original, cloner) {
      name = original.name;
      log = cloner.Clone(original.log);
      results = cloner.Clone(original.results);
      ensembles = (ItemList<RegressionEnsembleModel>) original.Ensembles.Clone(cloner);
      datastream = (Datastream) original.Datastream.Clone(cloner);
      runsCounter = original.runsCounter;
      runs = cloner.Clone(original.runs);
      Initialize();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new DatastreamAnalysisOptimizer(this, cloner);
    }

    private void Initialize() {
      if (runs != null) RegisterRunsEvents();
    }
    #endregion

    #region control actions

    public override void Prepare() {
      if (ensembles == null || datastream == null) return;
      Prepare(true);
    }

    public void Prepare(bool clearRuns) {
      if (ensembles == null || datastream == null) return;
      base.Prepare();
      if (clearRuns) results.Clear();
      OnPrepared();
    }

    public override void Start() {
      if (ensembles == null || datastream == null) return;
      base.Start();
      cancellationTokenSource = new CancellationTokenSource();
      stopPending = false;

      Task task = Task.Factory.StartNew(Run, cancellationTokenSource.Token, cancellationTokenSource.Token);
      task.ContinueWith(t => {
        try {
          t.Wait();
        }
        catch (AggregateException ex) {
          try {
            ex.Flatten().Handle(x => x is OperationCanceledException);
          } catch (AggregateException remaining) {
            if(remaining.InnerExceptions.Count == 1) OnExceptionOccurred(remaining.InnerExceptions[0]);
            else OnExceptionOccurred(remaining);
          }
        }
        cancellationTokenSource.Dispose();
        cancellationTokenSource = null;

        // handle stop/pause
        if(stopPending) { }
        if(!Datastream.SlidingWindowMovementPossible) OnStopped();
        else OnPaused();
      });

    }

    public override void Pause() {
      if (ensembles == null || datastream == null) return;
      base.Pause();
      cancellationTokenSource.Cancel();
    }

    public override void Stop() {
      if (ensembles == null || datastream == null) return;
      base.Stop();
      if (ExecutionState == ExecutionState.Paused) {
        OnStopped();
      } else {
        stopPending = true;
        cancellationTokenSource.Cancel();
      }
    }

    protected override void OnPrepared() {
      Log.LogMessage("Datastream analysis prepared");
      base.OnPrepared();
    }

    protected override void OnStarted() {
      Log.LogMessage("Datastream analysis started");
      base.OnStarted();
    }

    protected override void OnPaused() {
      Log.LogMessage("Datastream analysis paused");
      base.OnPaused();
    }

    protected override void OnStopped() {
      Log.LogMessage("Datastream analysis stopped");
      base.OnStopped();
    }

    #endregion


    private void Run(object state) {
      CancellationToken cancellationToken = (CancellationToken) state;
      OnStarted();
      lastUpdateTime = DateTime.UtcNow;
      System.Timers.Timer timer = new System.Timers.Timer(250);
      timer.AutoReset = true;
      timer.Elapsed += new System.Timers.ElapsedEventHandler(timer_Elapsed);
      timer.Start();

      try {
        Run(cancellationToken);
      }
      finally {
        timer.Elapsed -= new System.Timers.ElapsedEventHandler(timer_Elapsed);
        timer.Stop();
        ExecutionTime += DateTime.UtcNow - lastUpdateTime;
      }

      cancellationToken.ThrowIfCancellationRequested();
    }

    protected void Run(CancellationToken cancellationToken) {

      // setup results
      var slidingWindowMovements = new IntValue(0);
      Results.Add(new Result("Sliding Window Movements", slidingWindowMovements));

      var qtable = new DataTable("Qualities");
      qtable.Rows.Add(new DataRow("R²"));
      qtable.Rows.Add(new DataRow("Pearson"));
      Results.Add(new Result("Qualities", qtable));

      var curLoss = new DoubleValue();
      Results.Add(new Result("R²", curLoss));
      Results.Add(new Result("Best Quality", curLoss));


      // init algorithm
      var problemData = Datastream.ProblemData;
      var targetVarName = problemData.TargetVariable;
      var activeVariables = problemData.AllowedInputVariables;
      Random rnd = new Random();

      try {
        while (Datastream.SlidingWindowMovementPossible) {
          cancellationToken.ThrowIfCancellationRequested();

          Task.Delay(Datastream.SlidingWindowMovementInterval.Value);

          Datastream.MoveSlidingWindow();

          // TODO do the window evaluation

          // TODO: collect results and display them

          curLoss.Value = rnd.Next(100);
          qtable.Rows["R²"].Values.Add(curLoss.Value);
          qtable.Rows["Pearson"].Values.Add(curLoss.Value % 10);

          slidingWindowMovements.Value++;
        }

        // TODO: collect runs (c.f. goal seeking)
      }
      catch (Exception ex) {
        if (ex is ArgumentOutOfRangeException) throw ex;
        if (ex is OperationCanceledException) throw ex;
      }
      finally {
        // reset everything
        //Prepare(true);
      }
    }

    private void timer_Elapsed(object sender, System.Timers.ElapsedEventArgs e) {
      System.Timers.Timer timer = (System.Timers.Timer)sender;
      timer.Enabled = false;
      DateTime now = DateTime.UtcNow;
      ExecutionTime += now - lastUpdateTime;
      lastUpdateTime = now;
      timer.Enabled = true;
    }

    public void CollectResultValues(IDictionary<string, IItem> values) {
      values.Add("Execution Time", new TimeSpanValue(ExecutionTime));
      Results.CollectResultValues(values);
    }

    #region events

    public EventHandler EnsemblesChanged;
    public EventHandler DatastreamChanged;

    protected virtual void DeregisterRunsEvents() {
      runs.CollectionReset -= new CollectionItemsChangedEventHandler<IRun>(Runs_CollectionReset);
    }

    protected virtual void RegisterRunsEvents() {
      runs.CollectionReset += new CollectionItemsChangedEventHandler<IRun>(Runs_CollectionReset);
    }

    protected virtual void Runs_CollectionReset(object sender, CollectionItemsChangedEventArgs<IRun> e) {
      runsCounter = runs.Count;
    }

    #region event handling

    private void OnEnsemblesChanged() {
      var changed = EnsemblesChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    private void OnDatastreamChanged() {
      var changed = DatastreamChanged;
      if (changed != null)
        changed(this, EventArgs.Empty);
    }

    #endregion

    #endregion

    #region nameditem

    [Storable] protected string name;

    /// <inheritdoc/>
    /// <remarks>Calls <see cref="OnNameChanging"/> and also <see cref="OnNameChanged"/> 
    /// eventually in the setter.</remarks>
    public string Name {
      get { return name; }
      set {
        if (!CanChangeName) throw new NotSupportedException("Name cannot be changed.");
        if (!(name.Equals(value) || (value == null) && (name == string.Empty))) {
          CancelEventArgs<string> e = value == null
            ? new CancelEventArgs<string>(string.Empty)
            : new CancelEventArgs<string>(value);
          OnNameChanging(e);
          if (!e.Cancel) {
            name = value == null ? string.Empty : value;
            OnNameChanged();
          }
        }
      }
    }

    public virtual bool CanChangeName {
      get { return true; }
    }

    [Storable] protected string description;

    public string Description {
      get { return description; }
      set {
        if (!CanChangeDescription) throw new NotSupportedException("Description cannot be changed.");
        if (!(description.Equals(value) || (value == null) && (description == string.Empty))) {
          description = value == null ? string.Empty : value;
          OnDescriptionChanged();
        }
      }
    }

    public virtual bool CanChangeDescription {
      get { return true; }
    }

    /// <summary>
    /// Gets the string representation of the current instance in the format: <c>Name: [null|Value]</c>.
    /// </summary>
    /// <returns>The current instance as a string.</returns>
    public override string ToString() {
      return Name;
    }

    /// <inheritdoc/>
    public event EventHandler<CancelEventArgs<string>> NameChanging;

    /// <summary>
    /// Fires a new <c>NameChanging</c> event.
    /// </summary>
    /// <param name="e">The event arguments of the changing.</param>
    protected virtual void OnNameChanging(CancelEventArgs<string> e) {
      var handler = NameChanging;
      if (handler != null) handler(this, e);
    }

    /// <inheritdoc/>
    public event EventHandler NameChanged;

    /// <summary>
    /// Fires a new <c>NameChanged</c> event.
    /// </summary>
    /// <remarks>Calls <see cref="ItemBase.OnChanged"/>.</remarks>
    protected virtual void OnNameChanged() {
      var handler = NameChanged;
      if (handler != null) handler(this, EventArgs.Empty);
      OnToStringChanged();
    }

    /// <inheritdoc/>
    public event EventHandler DescriptionChanged;

    /// <summary>
    /// Fires a new <c>DescriptionChanged</c> event.
    /// </summary>
    /// <remarks>Calls <see cref="ItemBase.OnChanged"/>.</remarks>
    protected virtual void OnDescriptionChanged() {
      var handler = DescriptionChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    #endregion nameditem
  }
}